# (C) Copyright 2020 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Created on Nov 12, 2020

from sklearn.linear_model import LogisticRegression, LinearRegression
from causallib.evaluation import PropensityEvaluator, OutcomeEvaluator
from causallib.estimation import AIPW, IPW, StratifiedStandardization
from causallib.datasets import load_nhefs
import unittest
import pandas as pd
from sklearn.utils import Bunch
import matplotlib
matplotlib.use('Agg')


class TestPlots(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.data = load_nhefs()
        ipw = IPW(LogisticRegression(solver="liblinear"), clip_min=0.05, clip_max=0.95)
        std = StratifiedStandardization(LinearRegression())
        self.dr = AIPW(std, ipw)
        self.dr.fit(self.data.X, self.data.a, self.data.y)
        self.prp_evaluator = PropensityEvaluator(self.dr.weight_model)
        self.out_evaluator = OutcomeEvaluator(self.dr.outcome_model)

    def propensity_plot_by_name(self, test_names, alternate_a=None):
        a = self.data.a if alternate_a is None else alternate_a
        nhefs_plots = self.prp_evaluator.evaluate_simple(
            self.data.X, a, self.data.y, plots=test_names)
        [self.assertIsNotNone(x) for x in nhefs_plots.plots.values()]
        return True

    def outcome_plot_by_name(self, test_names):
        nhefs_plots = self.out_evaluator.evaluate_simple(
            self.data.X, self.data.a, self.data.y, plots=test_names)
        [self.assertIsNotNone(x) for x in nhefs_plots.plots.values()]
        return True

    def propensity_plot_multiple_a(self, test_names):
        self.assertTrue(self.propensity_plot_by_name(test_names, alternate_a=self.data.a.astype(int)))
        self.assertTrue(self.propensity_plot_by_name(test_names, alternate_a=self.data.a.astype(float)))
        # self.assertTrue(self.propensity_plot_by_name(test_names, alternate_a=self.data.a.astype(str).factorize()))


    def test_weight_distribution_plot(self):
        self.propensity_plot_multiple_a(["weight_distribution"])

    def test_propensity_roc_plots(self):
        self.propensity_plot_multiple_a(['roc_curve'])

    def test_precision_plots(self):
        self.propensity_plot_multiple_a(['precision'])

    def test_precision_plots(self):
        self.propensity_plot_multiple_a(['covariate_balance_love'])

    def test_propensity_multiple_plots(self):
        self.propensity_plot_multiple_a(['roc_curve', 'covariate_balance_love'])

    def test_accuracy_plot(self):
        self.assertTrue(self.outcome_plot_by_name(
            ["common_support", "continuous_accuracy"]))

# todo: add more tests (including ones that raise exceptions). No point in doing this right now since a major refactoring for the plots is ongoing



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Dec 25, 2018

"""

import abc
import warnings
from copy import deepcopy

import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.model_selection import StratifiedKFold

from .plots import get_subplots, plot_calibration_folds, plot_roc_curve_folds, plot_precision_recall_curve_folds


# TODO: How doubly robust fits in to show both weight and outcome model (at least show the plots on the same figure?)

class EvaluationResults:
    """Data structure to hold evaluation results"""

    def __init__(self, scores, plots, models):
        """
        Args:
            scores (pd.DataFrame or WeightEvaluatorScores):
            plots (dict): plot names to plot axes mapping.
                          Can be further nested by phase ("train"/"valid")
            models (list[WeightEstimator or IndividualOutcomeEstimator]): Models trained during evaluation
        """
        self.scores = scores
        self.plots = plots
        self.models = models


class BaseEvaluator:
    _numerical_classification_metrics = {"accuracy": metrics.accuracy_score,
                                         "precision": metrics.precision_score,
                                         "recall": metrics.recall_score,
                                         "f1": metrics.f1_score,
                                         "roc_auc": metrics.roc_auc_score,
                                         "avg_precision": metrics.average_precision_score,
                                         "hinge": metrics.hinge_loss,
                                         "matthews": metrics.matthews_corrcoef,
                                         "0/1": metrics.zero_one_loss,
                                         "brier": metrics.brier_score_loss}
    _nonnumerical_classification_metrics = {"confusion_matrix": metrics.confusion_matrix,
                                            "roc_curve": metrics.roc_curve,
                                            "pr_curve": metrics.precision_recall_curve}
    _classification_metrics = {**_numerical_classification_metrics, **_nonnumerical_classification_metrics}

    _regression_metrics = {"expvar": metrics.explained_variance_score,
                           "mae": metrics.mean_absolute_error,
                           "mse": metrics.mean_squared_error,
                           "msle": metrics.mean_squared_log_error,
                           # Allow mdae receive sample_weight argument but ignore it. This unifies the interface:
                           "mdae": lambda y_true, y_pred, **kwargs: metrics.median_absolute_error(y_true, y_pred),
                           "r2": metrics.r2_score}

    def __init__(self, estimator):
        """

        Args:
            estimator (causallib.estimation.base_weight.WeightEstimator | causallib.estimation.base_estimator.IndividualOutcomeEstimator):
        """
        self.estimator = estimator
        self._plot_functions = {"roc_curve": plot_roc_curve_folds,
                                "pr_curve": plot_precision_recall_curve_folds,
                                "calibration": plot_calibration_folds}

    def score_binary_prediction(self, y_true, y_pred_proba=None, y_pred=None, sample_weight=None,
                                metrics_to_evaluate=None, only_numeric_metric=True):
        """Evaluates a binary prediction against true labels.

        Args:
            y_true (pd.Series): True labels
            y_pred_proba (pd.Series): continuous output of predictor (as in `predict_proba` or `decision_function`).
            y_pred (pd.Series): label (i.e., categories, decisions) predictions.
            sample_weight (pd.Series | None): weight of each sample.
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.
            only_numeric_metric (bool): If metrics_to_evaluate not provided and default is used, whether to use only
                                        numerical metrics (non-numerical are for example roc_curve, that returns vectors
                                        and not scalars).
                                        Ignored if metrics_to_evaluate is provided

        Returns:
            pd.Series: name of metric as index and the evaluated score as value.
        """
        if metrics_to_evaluate is None:
            metrics_to_evaluate = self._numerical_classification_metrics if only_numeric_metric \
                else self._classification_metrics
        scores = {}
        for metric_name, metric_func in metrics_to_evaluate.items():
            if metric_name in {"hinge", "brier", "roc_curve", "roc_auc", "pr_curve", "avg_precision"}:
                prediction = y_pred_proba
            else:
                prediction = y_pred

            if prediction is None:
                continue

            try:
                scores[metric_name] = metric_func(y_true, prediction, sample_weight=sample_weight)
            except ValueError as v:  # if y_true has single value
                warnings.warn('metric {} could not be evaluated'.format(metric_name))
                warnings.warn(str(v))
                scores[metric_name] = np.nan

        dtype = float if all([np.isscalar(score) for score in scores.values()]) else np.dtype(object)
        return pd.Series(scores, dtype=dtype)

    def score_regression_prediction(self, y_true, y_pred, sample_weight=None, metrics_to_evaluate=None):
        """Evaluates continuous prediction against true labels

        Args:
            y_true (pd.Series): True label.
            y_pred (pd.Series): Predictions.
            sample_weight (pd.Series | None): weight for each sample.
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.

        Returns:
            pd.Series: name of metric as index and the evaluated score as value.
        """
        metrics_to_evaluate = metrics_to_evaluate or self._regression_metrics
        scores = {}
        for metric_name, metric_func in metrics_to_evaluate.items():
            try:
                scores[metric_name] = metric_func(y_true, y_pred, sample_weight=sample_weight)
            except ValueError as v:
                scores[metric_name] = np.nan
                warnings.warn('While evaluating ' + metric_name + ': ' + str(v))
        return pd.Series(scores)

    def evaluate_simple(self, X, a, y, metrics_to_evaluate=None, plots=None):
        """Evaluate model on the provided data

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Treatment assignment.
            y (pd.Series): Outcome.
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.
            plots (list[str] | None): list of plots to make. If None, none are generated.

        Returns:
            EvaluationResults
        """
        # simple evaluation without cross validation on the provided data
        # (can be to test the model on its train data or on new data

        phases = ["train"]  # dummy phase
        cv = pd.RangeIndex(start=0, stop=X.shape[0])  # All DataFrame rows when using iloc
        cv = [(cv, cv)]  # wrap in a tuple format compatible with sklearn's cv output
        results = self.evaluate_cv(X, a, y, cv=cv, refit=False, phases=phases,
                                   metrics_to_evaluate=metrics_to_evaluate, plots=plots)

        # Remove redundant information accumulated due to the use of cross-validation process
        results.plots = results.plots.pop('train', results.plots)
        results.models = results.models[0]
        scores = [results.scores] if isinstance(results.scores, pd.DataFrame) else results.scores
        for score in scores:
            score.reset_index(level=["phase", "fold"], drop=True, inplace=True)

        return results

    def evaluate_bootstrap(self, X, a, y, n_bootstrap, n_samples=None, replace=True,
                           refit=False, metrics_to_evaluate=None):
        """Evaluate model on a bootstrap sample of the provided data

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Treatment assignment.
            y (pd.Series): Outcome.
            n_bootstrap (int): Number of bootstrap sample to create.
            n_samples (int | None): Number of samples to sample in each bootstrap sampling.
                                    If None - will use the number samples (first dimension) of the data.
            replace (bool): Whether to use sampling with replacements.
                            If False - n_samples (if provided) should be smaller than X.shape[0])
            refit (bool): Whether to refit the estimator on each bootstrap sample.
                          Can be computational intensive if n_bootstrap is large.
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.

        Returns:
            EvaluationResults
        """
        n_samples = n_samples or X.shape[0]
        # Evaluation using bootstrap
        phases = ["train"]  # dummy phase

        # Generate bootstrap sample:
        cv = []
        X_ilocs = pd.RangeIndex(start=0, stop=X.shape[0])  # All DataFrame rows when using iloc
        for i in range(n_bootstrap):
            # Get iloc positions of a bootstrap sample (sample the size of X with replacement):
            # idx = X.sample(n=X.shape[0], replace=True).index
            # idx = np.random.random_integers(low=0, high=X.shape[0], size=X.shape[0])
            idx = np.random.choice(X_ilocs, size=n_samples, replace=replace)
            cv.append((idx, idx))  # wrap in a tuple format compatible with sklearn's cv output

        results = self.evaluate_cv(X, a, y, cv=cv, refit=refit, phases=phases,
                                   metrics_to_evaluate=metrics_to_evaluate, plots=None)

        # Remove redundant information accumulated due to the use of cross-validation process:
        results.models = results.models[0] if len(results.models) == 1 else results.models
        scores = [results.scores] if isinstance(results.scores, pd.DataFrame) else results.scores
        for score in scores:
            score.reset_index(level=["phase"], drop=True, inplace=True)
            score.index.rename("sample", "fold", inplace=True)
        return results

    def evaluate_cv(self, X, a, y, cv=None, kfold=None, refit=True, phases=("train", "valid"),
                    metrics_to_evaluate=None, plots=None):
        """Evaluate model in cross-validation of the provided data

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Treatment assignment.
            y (pd.Series): Outcome.
            cv (list[tuples] | generator[tuples]): list the number of folds containing tuples of indices
                                                   (train_idx, validation_idx) in an iloc manner (row number).
            kfold(sklearn.model_selection.BaseCrossValidator): Initialized fold object (e.g. KFold).
                                                               defaults to StratifiedKFold of 5 splits on treatment.
            refit (bool): Whether to refit the model on each fold.
            phases (list[str]): {["train", "valid"], ["train"], ["valid"]}.
                                Phases names to evaluate on - train ("train"), validation ("valid") or both.
                                'train' corresponds to cv[i][0] and 'valid' to  cv[i][1]
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.
            plots (list[str] | None): list of plots to make. If None, none are generated.

        Returns:
            EvaluationResults
        """
        # There's a need to have consistent splits for predicting, scoring and plotting.
        # If cv is a generator, it would be lost after after first use. if kfold has shuffle=True, it would be
        # inconsistent. In order to keep consistent reproducible folds across the process, we save them as a list.
        if cv is not None:
            cv = list(cv)  # if cv is generator it would listify it, if cv is already a list this is idempotent
        else:
            kfold = kfold or StratifiedKFold(n_splits=5)
            cv = list(kfold.split(X=X, y=a))

        predictions, models = self.predict_cv(X, a, y, cv, refit, phases)

        scores = self.score_cv(predictions, X, a, y, cv, metrics_to_evaluate)

        if plots is not None:
            plots = self.plot_cv(predictions, X, a, y, cv, plots)

        return_values = EvaluationResults(scores=scores,
                                          plots=plots if plots is not None else {},
                                          models=models if refit is True else [self.estimator])
        return return_values

    def predict_cv(self, X, a, y, cv, refit=True, phases=("train", "valid")):
        """Obtain predictions on the provided data in cross-validation

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Treatment assignment.
            y (pd.Series): Outcome.
            cv (list[tuples]): list the number of folds containing tuples of indices (train_idx, validation_idx)
            refit (bool): Whether to refit the model on each fold.
            phases (list[str]): {["train", "valid"], ["train"], ["valid"]}.
                                Phases names to evaluate on - train ("train"), validation ("valid") or both.
                                'train' corresponds to cv[i][0] and 'valid' to  cv[i][1]
        Returns:
            (dict[str, list], list): A two-tuple containing:

                * predictions: dictionary with keys being the phases provided and values are list the size of the number
                               of folds in cv and containing the output of the estimator on that corresponding fold.
                               For example, predictions["valid"][3] contains the prediction of the estimator on
                               untrained data of the third fold (i.e. validation set of the third fold)
                * models: list the size of the number of folds in cv containing of fitted estimator on the training data
                          of that fold.
        """

        predictions = {phase: [] for phase in phases}
        models = []
        for train_idx, valid_idx in cv:
            data = {"train": {"X": X.iloc[train_idx], "a": a.iloc[train_idx], "y": y.iloc[train_idx]},
                    "valid": {"X": X.iloc[valid_idx], "a": a.iloc[valid_idx], "y": y.iloc[valid_idx]}}
            # TODO: use dict-comprehension to map between phases[0] to cv[0] instead writing "train" explicitly

            if refit:
                self._estimator_fit(X=data["train"]["X"], a=data["train"]["a"], y=data["train"]["y"])

            for phase in phases:
                fold_prediction = self._estimator_predict(X=data[phase]["X"], a=data[phase]["a"])
                predictions[phase].append(fold_prediction)

            models.append(deepcopy(self.estimator))
        return predictions, models

    def score_cv(self, predictions, X, a, y, cv, metrics_to_evaluate=None):
        """Evaluate the prediction against the true data using evaluation score metrics.

        Args:
            predictions (dict[str, list]): the output of predict_cv.
            X (pd.DataFrame): Covariates.
            a (pd.Series): Treatment assignment.
            y (pd.Series): Outcome.
            cv (list[tuples]): list the number of folds containing tuples of indices (train_idx, validation_idx)
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.

        Returns:
            pd.DataFrame | WeightEvaluatorScores:
                DataFrame whose columns are different metrics and each row is a product of phase x fold x strata.
                WeightEvaluatorScores also has a covariate-balance result in a DataFrame.
        """
        phases = predictions.keys()
        scores = {phase: [] for phase in phases}
        for i, (train_idx, valid_idx) in enumerate(cv):
            data = {"train": {"X": X.iloc[train_idx], "a": a.iloc[train_idx], "y": y.iloc[train_idx]},
                    "valid": {"X": X.iloc[valid_idx], "a": a.iloc[valid_idx], "y": y.iloc[valid_idx]}}
            # TODO: use dict-comprehension to map between phases[0] to cv[0] instead writing "train" explicitly

            for phase in phases:
                X_fold, a_fold, y_fold = data[phase]["X"], data[phase]["a"], data[phase]["y"]
                prediction = predictions[phase][i]

                fold_scores = self.score_estimation(prediction, X_fold, a_fold, y_fold, metrics_to_evaluate)
                scores[phase].append(fold_scores)

        scores = self._combine_fold_scores(scores)
        return scores

    def plot_cv(self, predictions, X, a, y, cv, plots):
        """Plots prediction performance across different folds.

        Args:
            predictions (dict[str, list]): the output of predict_cv.
            X (pd.DataFrame): Covariates.
            a (pd.Series): Treatment assignment.
            y (pd.Series): Outcome.
            cv (list[tuples]): list the number of folds containing tuples of indices (train_idx, validation_idx)
            plots (list[str]): list of plots to make

        Returns:
            dict[str, dict[str, plt.Axes]]:
                {phase_name: {plot_name: plot_axes}}.
                 For example: {"train": {"roc_curve": plt.Axes, "calibration": plt.Axes}}
        """
        phases = predictions.keys()
        all_axes = {phase: {} for phase in phases}

        for phase in phases:
            phase_fig, phase_axes = get_subplots(len(plots))
            phase_axes = phase_axes.ravel()  # squeeze a vector out of the matrix-like structure of the returned fig.

            # Retrieve all indices of the different folds in the phase [idx_fold_1, idx_folds_2, ...]
            cv_idx_folds = [fold_idx[0] if phase == "train" else fold_idx[1] for fold_idx in cv]
            predictions_folds = predictions[phase]

            for i, plot_name in enumerate(plots):
                plot_data = self._get_data_for_plot(plot_name, predictions_folds, X, a, y, cv_idx_folds)
                # TODO: ^ consider _get_data_for_plot returning args (tuple) and kwargs (dictionary) which will be
                #       expanded when calling plot_func: plot_func(*plot_args, **plot_kwargs).
                #       This will allow more flexible specification of parameters by the caller
                #       (For example, Propensity Distribution with kde=True and Weight Distribution with kde=False)
                plot_func = self._plot_functions.get(plot_name)
                if plot_func is None or plot_data is None:
                    plot_ax = None
                else:
                    plot_ax = plot_func(*plot_data, cv=cv_idx_folds, ax=phase_axes[i])
                all_axes[phase][plot_name] = plot_ax
            phase_fig.suptitle("Evaluation on {} phase".format(phase))
        return all_axes

    @abc.abstractmethod
    def _estimator_fit(self, X, a, y):
        """Fit an estimator."""
        raise NotImplementedError

    @abc.abstractmethod
    def _estimator_predict(self, X, a):
        """Predict (weights, outcomes, etc. depending on the model).
        The output can be as flexible as desired, but score_estimation should know to handle it."""
        raise NotImplementedError

    @abc.abstractmethod
    def score_estimation(self, prediction, X, a_true, y_true, metrics_to_evaluate=None):
        """Should know how to handle the _estimator_predict output provided in `prediction`.
        Can utilize any of the true values provided (covariates `X`, treatment assignment `a` or outcome `y`)."""
        raise NotImplementedError

    @staticmethod
    def _combine_fold_scores(scores):
        """
        Combines scores of each phase and fold into a single object (DataFrame) of scores.

        Args:
            scores (dict[str, list[pd.DataFrame]]):
                scores of each fold of each phase. The structure is {phase_name: [fold_1_score, fold_2_score...]}.
                Where phase_name is usually "train" or "valid", and each fold_i_score is a DataFrame which columns are
                evaluation metrics and rows are results of that metrics in that fold.

        Returns:
            pd.DataFrame: Row-concatenated DataFrame with MultiIndex accounting for the concatenated folds and phases.
        """
        # Concatenate the scores from list of folds to DataFrame with rows as folds, keeping it by different phases:
        scores = {phase: pd.concat(scores_fold, axis="index", keys=range(len(scores_fold)), names=["fold"])
                  for phase, scores_fold in scores.items()}
        # Concatenate the train/validation DataFrame scores into DataFrame with rows as phases:
        scores = pd.concat(scores, axis="index", names=["phase"])
        return scores

    @abc.abstractmethod
    def _get_data_for_plot(self, plot_name, folds_predictions, X, a, y, cv):
        """Return a tuple containing the relevant data needed for the specific plot provided in `plot_name`"""
        raise NotImplementedError

    # Calculating ROC/PR curves:
    def _calculate_roc_curve_data(self, folds_predictions, targets, stratify_by=None):
        """Calculates ROC curve on the folds

        Args:
            folds_predictions (list[WeightEvaluatorPredictions | OutcomeEvaluatorPredictions]):
                list of the predictions, each entry correspond to a fold.
            targets (pd.Series): True labels.
            stratify_by (pd.Series): A vector (mostly, treatment assignment) to perform groupby with.

        Returns:
            dict[str, list[np.ndarray]]: Keys being "FPR", "TPR" and "AUC" (ROC metrics) and values are a list the size
                                         of number of folds with the evaluation of each fold.
        """
        curve_data = self._calculate_curve_data(folds_predictions, targets,
                                                metrics.roc_curve, metrics.roc_auc_score,
                                                stratify_by=stratify_by)
        for curve_name in curve_data.keys():
            curve_data[curve_name]["FPR"] = curve_data[curve_name].pop("first_ret_value")
            curve_data[curve_name]["TPR"] = curve_data[curve_name].pop("second_ret_value")
            curve_data[curve_name]["AUC"] = curve_data[curve_name].pop("area")
        return curve_data

    def _calculate_pr_curve_data(self, folds_predictions, targets, stratify_by=None):
        """Calculates precision-recall curve on the folds

        Args:
            folds_predictions (list[WeightEvaluatorPredictions | OutcomeEvaluatorPredictions]):
                list of the predictions, each entry correspond to a fold.
            targets (pd.Series): True labels.
            stratify_by (pd.Series): A vector (mostly, treatment assignment) to perform groupby with.

        Returns:
            dict[str, list[np.ndarray]]: Keys being "Precision", "Recall" and "AP" (PR metrics) and values are a list
                                         the size of number of folds with the evaluation of each fold.
                                         Additional "prevalence" key, with positive-label prevalence is added (to be
                                         used by the chance curve).
        """
        curve_data = self._calculate_curve_data(folds_predictions, targets,
                                                metrics.precision_recall_curve,
                                                metrics.average_precision_score,
                                                stratify_by=stratify_by)
        for curve_name in curve_data.keys():
            curve_data[curve_name]["Precision"] = curve_data[curve_name].pop("first_ret_value")
            curve_data[curve_name]["Recall"] = curve_data[curve_name].pop("second_ret_value")
            curve_data[curve_name]["AP"] = curve_data[curve_name].pop("area")
        curve_data["prevalence"] = targets.value_counts(normalize=True).loc[targets.max()]
        return curve_data

    @abc.abstractmethod
    def _calculate_curve_data(self, folds_predictions, targets, curve_metric, area_metric, stratify_by=None):
        """Given a list of predictions (the output of _estimator_predict by folds)
        and a vector of targets. extract (if needed)the relevant parts in each fold prediction
        and apply the curve_metric and area_metric.
        The output is nested dict: first level is the curve name and inner level dict has
        `first_ret_value` `first_ret_value`, `Thresholds` and `area` of that curve` in list
        with each entry corresponding to each fold."""
        raise NotImplementedError

    @staticmethod
    def _calculate_performance_curve_data_on_folds(folds_predictions, folds_targets, sample_weights=None,
                                                   area_metric=metrics.roc_auc_score, curve_metric=metrics.roc_curve,
                                                   pos_label=None):
        """Calculates performance curves (either ROC or precision-recall) of the predictions across folds.

        Args:
            folds_predictions (list[pd.Series]): Score prediction (as in continuous output of classifier,
                                                 `predict_proba` or `decision_function`) for every fold.
            folds_targets (list[pd.Series]): True labels for every fold.
            sample_weights (list[pd.Series] | None): weight for each sample for every fold.
            area_metric (callable): Performance metric of the area under the curve.
            curve_metric (callable): Performance metric returning 3 output vectors - metric1, metric2 and thresholds.
                                    Where metric1 and metric2 depict the curve when plotted on x-axis and y-axis.
            pos_label: What label in `targets` is considered the positive label.

        Returns:
            (list[np.ndarray], list[np.ndarray], list[np.ndarray], list[float]):
             For every fold, the calculated metric1 and metric2 (the curves), the thresholds and the area calculations.
        """
        sample_weights = [None] * len(folds_predictions) if sample_weights is None else sample_weights
        # Scikit-learn precision_recall_curve and roc_curve do not return values in a consistent way.
        # Namely, roc_curve returns `fpr`, `tpr`, which correspond to x_axis, y_axis,
        # whereas precision_recall_curve returns `precision`, `recall`, which correspond to y_axis, x_axis.
        # That's why this function will return the values the same order as the Scikit's curves, and leave it up to the
        # caller to put labels on what those return values actually are (specifically, whether they're x_axis or y-axis)
        first_ret_folds, second_ret_folds, threshold_folds, area_folds = [], [], [], []
        for fold_prediction, fold_target, fold_weights in zip(folds_predictions, folds_targets, sample_weights):
            first_ret_fold, second_ret_fold, threshold_fold = curve_metric(fold_target, fold_prediction,
                                                                           pos_label=pos_label,
                                                                           sample_weight=fold_weights)
            try:
                area_fold = area_metric(fold_target, fold_prediction, sample_weight=fold_weights)
            except ValueError as v:  # AUC cannot be evaluated if targets are constant
                warnings.warn('metric {} could not be evaluated'.format(area_metric.__name__))
                warnings.warn(str(v))
                area_fold = np.nan

            first_ret_folds.append(first_ret_fold)
            second_ret_folds.append(second_ret_fold)
            threshold_folds.append(threshold_fold)
            area_folds.append(area_fold)
        return area_folds, first_ret_folds, second_ret_folds, threshold_folds



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Dec 25, 2018

"""
from collections import namedtuple

import numpy as np
import pandas as pd

from .evaluator import BaseEvaluator
from .plots import plot_mean_features_imbalance_slope_folds, plot_mean_features_imbalance_love_folds
from .plots import plot_propensity_score_distribution_folds
from ..estimation.base_weight import WeightEstimator, PropensityEstimator
from ..utils.stat_utils import calc_weighted_standardized_mean_differences, calc_weighted_ks2samp, robust_lookup


# TODO: decide what implementation stays - the one with the '2' suffix or the one without.
#       The one with is based on matrix input and does all the vector extraction by itself.
#       The one without is simpler one the receive the vectors already (more general, as not all models may have matrix.


# ################ #
# Weight Evaluator #
# ################ #

class WeightEvaluatorPredictions:
    """Data structure to hold weight-model predictions"""

    def __init__(self, weight_by_treatment_assignment, weight_for_being_treated, treatment_assignment_pred):
        self.weight_by_treatment_assignment = weight_by_treatment_assignment
        self.weight_for_being_treated = weight_for_being_treated
        self.treatment_assignment_pred = treatment_assignment_pred


class WeightEvaluatorPredictions2:
    """Data structure to hold weight-model predictions"""

    def __init__(self, weight_matrix, treatment_assignment, treatment_assignment_prediction=None):
        self.weight_matrix = weight_matrix
        self._treatment_assignment = treatment_assignment
        self.treatment_assignment_prediction = treatment_assignment_prediction

    @property
    def weight_by_treatment_assignment(self):
        weight_by_treatment_assignment = self._extract_vector_from_matrix(self.weight_matrix,
                                                                          self._treatment_assignment)
        return weight_by_treatment_assignment

    @property
    def weight_for_being_treated(self):
        weight_for_being_treated = self._extract_vector_from_matrix(self.weight_matrix,
                                                                    self._treatment_assignment.max())
        return weight_for_being_treated

    @staticmethod
    def _extract_vector_from_matrix(matrix, value):
        if np.isscalar(value):
            vector = matrix[value]
        else:
            # vector = robust_lookup(matrix, value)
            vector = matrix.lookup(value.index, value)
            vector = pd.Series(vector, index=value.index)
        return vector


WeightEvaluatorScores = namedtuple("WeightEvaluatorScores", ["prediction_scores", "covariate_balance"])


class WeightEvaluator(BaseEvaluator):
    def __init__(self, estimator):
        """
        Args:
            estimator (WeightEstimator):
        """
        if not isinstance(estimator, WeightEstimator):
            raise TypeError("WeightEvaluator should be initialized with WeightEstimator, got ({}) instead."
                            .format(type(estimator)))
        super(WeightEvaluator, self).__init__(estimator)
        self._plot_functions.update({"weight_distribution": plot_propensity_score_distribution_folds,
                                     "covariate_balance_love": plot_mean_features_imbalance_love_folds,
                                     "covariate_balance_slope": plot_mean_features_imbalance_slope_folds})

    def _estimator_fit(self, X, a, y=None):
        """ Fit estimator. `y` is ignored.
        """
        self.estimator.fit(X=X, a=a)

    def _estimator_predict(self, X, a):
        """Predict on data.

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Target variable - treatment assignment

        Returns:
            WeightEvaluatorPredictions
        """
        weight_by_treatment_assignment = self.estimator.compute_weights(X, a, treatment_values=None,
                                                                        use_stabilized=False)
        weight_for_being_treated = self.estimator.compute_weights(X, a, treatment_values=a.max(),
                                                                  use_stabilized=False)
        treatment_assignment_pred = self.estimator.learner.predict(
            X)  # TODO: maybe add predict_label to interface instead
        treatment_assignment_pred = pd.Series(treatment_assignment_pred, index=X.index)

        prediction = WeightEvaluatorPredictions(weight_by_treatment_assignment,
                                                weight_for_being_treated,
                                                treatment_assignment_pred)
        return prediction

    def _estimator_predict2(self, X, a):
        """Predict on data"""
        weight_matrix = self.estimator.compute_weight_matrix(X, a, use_stabilized=False)
        treatment_assignment_pred = self.estimator.learner.predict(
            X)  # TODO: maybe add predict_label to interface instead
        treatment_assignment_pred = pd.Series(treatment_assignment_pred, index=X.index)

        fold_prediction = WeightEvaluatorPredictions2(weight_matrix, a, treatment_assignment_pred)
        return fold_prediction

    def score_estimation(self, prediction, X, a_true, y_true=None, metrics_to_evaluate=None):
        """Scores a prediction against true labels.

        Args:
            prediction (WeightEvaluatorPredictions): Prediction on the data.
            X (pd.DataFrame): Covariates.
            a_true (pd.Series): Target variable - treatment assignment
            y_true: *IGNORED*
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.

        Returns:
            WeightEvaluatorScores: Data-structure holding scores on the predictions
                                   and covariate balancing table ("table 1")
        """
        results = self._score_estimation(X, a_true,
                                         prediction.weight_for_being_treated,
                                         prediction.treatment_assignment_pred,
                                         prediction.weight_by_treatment_assignment,
                                         metrics_to_evaluate)
        return results

    def _score_estimation(self, X, targets, predict_scores, predict_assignment, predict_weights,
                          metrics_to_evaluate=None):
        """

        Args:
            X (pd.DataFrame): Covariates.
            targets (pd.Series): Target variable - true treatment assignment
            predict_scores (pd.Series): Continuous prediction of the treatment assignment
                                        (as is `predict_proba` or `decision_function`).
            predict_assignment (pd.Series): Class prediction of the treatment assignment
                                            (i.e. prediction of the assignment it self).
            predict_weights (pd.Series): The weights derived to balance between the treatment groups
                                         (here, called `targets`).
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.

        Returns:
            WeightEvaluatorScores: Data-structure holding scores on the predictions
                                   and covariate balancing table ("table 1")
        """
        prediction_scores = self.score_binary_prediction(y_true=targets,
                                                         y_pred_proba=predict_scores,
                                                         y_pred=predict_assignment,
                                                         metrics_to_evaluate=metrics_to_evaluate)
        # Convert single-dtype Series to a row in a DataFrame:
        prediction_scores = pd.DataFrame(prediction_scores).T
        # change dtype of each column to numerical if possible:
        prediction_scores = prediction_scores.apply(pd.to_numeric, errors="ignore")

        covariate_balance = calculate_covariate_balance(X, targets, predict_weights)

        results = WeightEvaluatorScores(prediction_scores, covariate_balance)
        return results

    def _combine_fold_scores(self, scores):
        # `scores` are provided as WeightEvaluatorScores object for each fold in each phase,
        # Namely, dict[list[WeightEvaluatorScores]], which in turn hold two DataFrames components.
        # In order to combine the underlying DataFrames into a multilevel DataFrame, one must first extract them from
        # the WeightEvaluatorScores object, into two separate components.

        # Extract the two components of WeightEvaluatorScores:
        prediction_scores = {phase: [fold_score.prediction_scores for fold_score in phase_scores]
                             for phase, phase_scores in scores.items()}
        covariate_balance = {phase: [fold_score.covariate_balance for fold_score in phase_scores]
                             for phase, phase_scores in scores.items()}

        # Combine the dict[list[DataFrames]] of each component into a multilevel DataFrame separately:
        prediction_scores = super(WeightEvaluator, self)._combine_fold_scores(prediction_scores)
        covariate_balance = super(WeightEvaluator, self)._combine_fold_scores(covariate_balance)
        # TODO: consider reordering the levels, such that the covariate will be the first one and then phase and fold
        # covariate_balance = covariate_balance.reorder_levels(["covariate", "phase", "fold"])

        # Create a new WeightEvaluatorScores object with the combined (i.e., multilevel DataFrame) results:
        scores = WeightEvaluatorScores(prediction_scores, covariate_balance)
        return scores

    def _get_data_for_plot(self, plot_name, folds_predictions, X, a, y, cv):
        """Retrieve the data needed for each provided plot.
        Plot interfaces are at the plots.py module.

        Args:
            plot_name (str): Plot name.
            folds_predictions (list[WeightEvaluatorPredictions]): Predictions for each fold.
            X (pd.DataFrame): Covariates.
            a (pd.Series): Target variable - treatment assignment
            y: *IGNORED*
            cv list[np.ndarray]: Indices (in iloc positions) of each fold.

        Returns:
            tuple: Plot data
        """
        if plot_name in {'weight_distribution'}:
            folds_predictions = [prediction.weight_for_being_treated for prediction in folds_predictions]
            return folds_predictions, a
        elif plot_name in {'roc_curve'}:
            curve_data = self._calculate_roc_curve_data(folds_predictions, a)
            return (curve_data,)
        elif plot_name in {'pr_curve'}:
            curve_data = self._calculate_pr_curve_data(folds_predictions, a)
            return (curve_data,)
        elif plot_name in {"covariate_balance_love", "covariate_balance_slope"}:
            distribution_distances = []
            for fold_prediction in folds_predictions:
                fold_w = fold_prediction.weight_by_treatment_assignment
                fold_X = X.loc[fold_w.index]
                fold_a = a.loc[fold_w.index]
                dist_dist = calculate_covariate_balance(fold_X, fold_a, fold_w)
                distribution_distances.append(dist_dist)
            return (distribution_distances,)
        else:
            return None

    def _calculate_curve_data(self, folds_predictions, targets, curve_metric, area_metric, **kwargs):
        """Calculate different performance (ROC or PR) curves

        Args:
            folds_predictions (list[WeightEvaluatorPredictions]): Predictions for each fold.
            targets (pd.Series): True labels
            curve_metric (callable): Performance metric returning 3 output vectors - metric1, metric2 and thresholds.
                                    Where metric1 and metric2 depict the curve when plotted on x-axis and y-axis.
            area_metric (callable): Performance metric of the area under the curve.
            **kwargs:

        Returns:
            dict[str, dict[str, list[np.ndarray]]]: Evaluation of the metric for each fold and for each curve.
                2 curves:
                    * "unweighted" (regular)
                    * "weighted" (weighted by weights of each sample (according to their assignment))
                On general: {curve_name: {metric1: [evaluation_fold_1, ...]}}.
                For example: {"weighted": {"FPR": [FPR_fold_1, FPR_fold_2, FPR_fold3]}}
        """
        folds_sample_weights = {"unweighted": [None for _ in folds_predictions],
                                "weighted": [fold_predictions.weight_by_treatment_assignment
                                             for fold_predictions in folds_predictions]}
        folds_predictions = [fold_predictions.weight_for_being_treated for fold_predictions in folds_predictions]
        folds_targets = []
        for fold_predictions in folds_predictions:
            # Since this is weight estimator, which takes the inverse of a class prediction
            fold_targets = targets.loc[fold_predictions.index]
            fold_targets = fold_targets.replace({fold_targets.min(): fold_targets.max(),
                                                 fold_targets.max(): fold_targets.min()})
            folds_targets.append(fold_targets)

        curve_data = {}
        for curve_name, sample_weights in folds_sample_weights.items():
            area_folds, first_ret_folds, second_ret_folds, threshold_folds = self._calculate_performance_curve_data_on_folds(
                folds_predictions, folds_targets, sample_weights, area_metric, curve_metric)

            curve_data[curve_name] = {"first_ret_value": first_ret_folds,
                                      "second_ret_value": second_ret_folds,
                                      "Thresholds": threshold_folds, "area": area_folds}

        # Rename keys (as will be presented as curve labels in legend)
        curve_data["Weights"] = curve_data.pop("unweighted")
        curve_data["Weighted"] = curve_data.pop("weighted")
        return curve_data


# #################### #
# Propensity Evaluator #
# #################### #


class PropensityEvaluatorPredictions(WeightEvaluatorPredictions):
    """Data structure to hold propensity-model predictions"""

    def __init__(self, weight_by_treatment_assignment, weight_for_being_treated, treatment_assignment_pred,
                 propensity, propensity_by_treatment_assignment):
        super(PropensityEvaluatorPredictions, self).__init__(weight_by_treatment_assignment,
                                                             weight_for_being_treated,
                                                             treatment_assignment_pred)
        self.propensity = propensity
        self.propensity_by_treatment_assignment = propensity_by_treatment_assignment


class PropensityEvaluatorPredictions2(WeightEvaluatorPredictions2):
    """Data structure to hold propensity-model predictions"""

    def __init__(self, weight_matrix, propensity_matrix, treatment_assignment, treatment_assignment_prediction=None):
        super(PropensityEvaluatorPredictions2, self).__init__(weight_matrix, treatment_assignment,
                                                              treatment_assignment_prediction)
        self.propensity_matrix = propensity_matrix

    @property
    def propensity(self):
        propensity = self._extract_vector_from_matrix(self.propensity_matrix,
                                                      self._treatment_assignment)
        return propensity

    @property
    def propensity_by_treatment_assignment(self):
        # TODO: remove propensity_by_treatment if expected-ROC is not to be used.
        propensity_by_treatment_assignment = self._extract_vector_from_matrix(self.propensity_matrix,
                                                                              self._treatment_assignment.max())
        return propensity_by_treatment_assignment


class PropensityEvaluator(WeightEvaluator):
    def __init__(self, estimator):
        """
        Args:
            estimator (PropensityEstimator):
        """
        if not isinstance(estimator, PropensityEstimator):
            raise TypeError("PropensityEvaluator should be initialized with PropensityEstimator, got ({}) instead."
                            .format(type(estimator)))
        super(PropensityEvaluator, self).__init__(estimator)

    def _estimator_predict(self, X, a):
        """Predict on data.

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Target variable - treatment assignment

        Returns:
            PropensityEvaluatorPredictions
        """
        propensity = self.estimator.compute_propensity(X, a, treatment_values=a.max())
        propensity_by_treatment_assignment = self.estimator.compute_propensity_matrix(X)
        propensity_by_treatment_assignment = robust_lookup(propensity_by_treatment_assignment, a)

        weight_prediction = super(PropensityEvaluator, self)._estimator_predict(X, a)
        # Do not force stabilize=False as in WeightEvaluator:
        weight_by_treatment_assignment = self.estimator.compute_weights(X, a)
        prediction = PropensityEvaluatorPredictions(weight_by_treatment_assignment,
                                                    weight_prediction.weight_for_being_treated,
                                                    weight_prediction.treatment_assignment_pred,
                                                    propensity,
                                                    propensity_by_treatment_assignment)
        return prediction

    def _estimator_predict2(self, X, a):
        """Predict on data."""
        weight_prediction = super(PropensityEvaluator, self)._estimator_predict2(X, a)
        propensity_matrix = self.estimator.compute_propensity_matrix(X)
        fold_prediction = PropensityEvaluatorPredictions2(weight_prediction.weight_matrix, propensity_matrix,
                                                          a, weight_prediction.treatment_assignment_prediction)
        return fold_prediction

    def score_estimation(self, prediction, X, a_true, y_true=None, metrics_to_evaluate=None):
        """Scores a prediction against true labels.

        Args:
            prediction (PropensityEvaluatorPredictions): Prediction on the data.
            X (pd.DataFrame): Covariates.
            a_true (pd.Series): Target variable - treatment assignment
            y_true: *IGNORED*
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.

        Returns:
            WeightEvaluatorScores: Data-structure holding scores on the predictions
                                   and covariate balancing table ("table 1")
        """
        results = self._score_estimation(X, a_true,
                                         prediction.propensity,
                                         prediction.treatment_assignment_pred,
                                         prediction.weight_by_treatment_assignment,
                                         metrics_to_evaluate)
        return results

    def _get_data_for_plot(self, plot_name, folds_predictions, X, a, y, cv):
        """Retrieve the data needed for each provided plot.
        Plot interfaces are at the plots.py module.

        Args:
            plot_name (str): Plot name.
            folds_predictions (list[PropensityEvaluatorPredictions]): Predictions for each fold.
            X (pd.DataFrame): Covariates.
            a (pd.Series): Target variable - treatment assignment
            y: *IGNORED*
            cv list[np.ndarray]: Indices (in iloc positions) of each fold.

        Returns:
            tuple: Plot data
        """
        if plot_name in {'weight_distribution'}:
            folds_predictions = [prediction.propensity for prediction in folds_predictions]
            return folds_predictions, a
        elif plot_name in {'calibration'}:
            folds_predictions = [prediction.propensity for prediction in folds_predictions]
            return folds_predictions, a
        else:
            # Common plots are implemented at top-most level possible.
            # Plot might be implemented by WeightEvaluator:
            return super(PropensityEvaluator, self)._get_data_for_plot(plot_name, folds_predictions, X, a, y, cv)

    def _calculate_curve_data(self, curves_folds_predictions, targets, curve_metric, area_metric, **kwargs):
        """Calculate different performance (ROC or PR) curves

        Args:
            curves_folds_predictions (list[PropensityEvaluatorPredictions]): Predictions for each fold.
            targets (pd.Series): True labels
            curve_metric (callable): Performance metric returning 3 output vectors - metric1, metric2 and thresholds.
                                    Where metric1 and metric2 depict the curve when plotted on x-axis and y-axis.
            area_metric (callable): Performance metric of the area under the curve.
            **kwargs:

        Returns:
            dict[str, dict[str, list[np.ndarray]]]: Evaluation of the metric for each fold and for each curve.
                3 curves:
                    * "unweighted" (regular)
                    * "weighted" (weighted by inverse propensity)
                    * "expected" (duplicated population, weighted by propensity)
                On general: {curve_name: {metric1: [evaluation_fold_1, ...]}}.
                For example: {"weighted": {"FPR": [FPR_fold_1, FPR_fold_2, FPR_fold3]}}
        """
        curves_sample_weights = {"unweighted": [None for _ in curves_folds_predictions],
                                 "weighted": [fold_predictions.weight_by_treatment_assignment
                                              for fold_predictions in curves_folds_predictions],
                                 "expected": [fold_predictions.propensity.append(1 - fold_predictions.propensity)
                                              for fold_predictions in curves_folds_predictions]}
        curves_folds_targets = [targets.loc[fold_predictions.weight_by_treatment_assignment.index]
                                for fold_predictions in curves_folds_predictions]
        curves_folds_targets = {
            "unweighted": curves_folds_targets,
            "weighted": curves_folds_targets,
            "expected": [pd.Series(data=targets.max(), index=fold_predictions.propensity.index).append(
                pd.Series(data=targets.min(), index=fold_predictions.propensity.index))
                for fold_predictions in curves_folds_predictions]
        }
        curves_folds_predictions = {
            "unweighted": [fold_predictions.propensity for fold_predictions in curves_folds_predictions],
            "weighted": [fold_predictions.propensity for fold_predictions in curves_folds_predictions],
            "expected": [fold_predictions.propensity.append(fold_predictions.propensity)
                         for fold_predictions in curves_folds_predictions]
        }
        # Expected curve duplicates the population, basically concatenating so that:
        # prediction = [p, p], target = [1, 0], weights = [p, 1-p]

        curve_data = {}
        for curve_name in curves_sample_weights.keys():
            sample_weights = curves_sample_weights[curve_name]
            folds_targets = curves_folds_targets[curve_name]
            folds_predictions = curves_folds_predictions[curve_name]

            area_folds, first_ret_folds, second_ret_folds, threshold_folds = self._calculate_performance_curve_data_on_folds(
                folds_predictions, folds_targets, sample_weights, area_metric, curve_metric)

            curve_data[curve_name] = {"first_ret_value": first_ret_folds,
                                      "second_ret_value": second_ret_folds,
                                      "Thresholds": threshold_folds, "area": area_folds}

        # Rename keys (as will be presented as curve labels in legend)
        curve_data["Propensity"] = curve_data.pop("unweighted")
        curve_data["Weighted"] = curve_data.pop("weighted")
        curve_data["Expected"] = curve_data.pop("expected")
        return curve_data


# ################# #
# Covariate Balance #
# ################# #


DISTRIBUTION_DISTANCE_METRICS = {"smd": lambda x, y, wx, wy: calc_weighted_standardized_mean_differences(x, y, wx, wy),
                                 "abs_smd": lambda x, y, wx, wy:
                                 abs(calc_weighted_standardized_mean_differences(x, y, wx, wy)),
                                 "ks": lambda x, y, wx, wy: calc_weighted_ks2samp(x, y, wx, wy)}


def calculate_covariate_balance(X, a, w, metric="abs_smd"):
    """Calculate covariate balance table ("table 1")

    Args:
        X (pd.DataFrame): Covariates.
        a (pd.Series): Group assignment of each sample.
        w (pd.Series): sample weights for balancing between groups in `a`.
        metric (str | callable): Either a key from DISTRIBUTION_DISTANCE_METRICS or a metric with the signature
                                 weighted_distance(x, y, wx, wy) calculating distance between the weighted sample x
                                 and weighted sample y (weights by wx and wy respectively).

    Returns:
        pd.DataFrame: index are covariate names (columns) from X, and columns are "weighted" / "unweighted" results
                      of applying `metric` on each covariate to compare the two groups.
    """
    treatment_values = np.sort(np.unique(a))
    results = {}
    for treatment_value in treatment_values:
        distribution_distance_of_cur_treatment = pd.DataFrame(index=X.columns, columns=["weighted", "unweighted"],
                                                              dtype=float)
        for col_name, col_data in X.items():
            weighted_distance = calculate_distribution_distance_for_single_feature(col_data, w, a,
                                                                                   treatment_value, metric)
            unweighted_distance = calculate_distribution_distance_for_single_feature(col_data,
                                                                                     pd.Series(1, index=w.index), a,
                                                                                     treatment_value, metric)
            distribution_distance_of_cur_treatment.loc[col_name, ["weighted", "unweighted"]] = \
                [weighted_distance, unweighted_distance]
        results[treatment_value] = distribution_distance_of_cur_treatment
    results = pd.concat(results, axis="columns", names=[a.name or "a", metric])  # type: pd.DataFrame
    results.index.name = "covariate"
    if len(treatment_values) == 2:
        # In case there's only two treatments, the results for treatment_value==0 and treatment_value==0 are identical.
        # Therefore, we can get rid from one of them.
        # Here we keep the results for the treated-group (noted by maximal treatment value, probably 1):
        results = results.xs(treatment_values.max(), axis="columns", level=0)
    # TODO: is there a neat expansion for multi-treatment case? maybe not current_treatment vs. the rest.
    return results


def calculate_distribution_distance_for_single_feature(x, w, a, group_level, metric="abs_smd"):
    """

    Args:
        x (pd.Series): A single feature to check balancing.
        a (pd.Series): Group assignment of each sample.
        w (pd.Series): sample weights for balancing between groups in `a`.
        group_level: Value from `a` in order to divide the sample into one vs. rest.
        metric (str | callable): Either a key from DISTRIBUTION_DISTANCE_METRICS or a metric with the signature
                                 weighted_distance(x, y, wx, wy) calculating distance between the weighted sample x
                                 and weighted sample y (weights by wx and wy respectively).

    Returns:
        float: weighted distance between the samples assigned to `group_level` and the rest of the samples.
    """
    if not callable(metric):
        metric = DISTRIBUTION_DISTANCE_METRICS[metric]
    cur_treated_mask = a == group_level
    x_treated = x.loc[cur_treated_mask]
    w_treated = w.loc[cur_treated_mask]
    x_untreated = x.loc[~cur_treated_mask]
    w_untreated = w.loc[~cur_treated_mask]
    distribution_distance = metric(x_treated, x_untreated, w_treated, w_untreated)
    return distribution_distance



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Dec 25, 2018

"""
import warnings

import numpy as np
import pandas as pd

from .evaluator import BaseEvaluator
from .plots import plot_counterfactual_common_support_folds, plot_continuous_prediction_accuracy_folds, \
    plot_residual_folds
from ..estimation.base_estimator import IndividualOutcomeEstimator
from ..utils.stat_utils import is_vector_binary
from ..utils.stat_utils import robust_lookup


class OutcomeEvaluatorPredictions:
    """Data structure to hold outcome-model predictions"""

    def __init__(self, prediction, prediction_event_prob=None):
        self.prediction = prediction
        self.prediction_event_prob = prediction_event_prob


class OutcomeEvaluator(BaseEvaluator):
    def __init__(self, estimator):
        """
        Args:
            estimator (IndividualOutcomeEstimator):
        """
        if not isinstance(estimator, IndividualOutcomeEstimator):
            raise TypeError("OutcomeEvaluator should be initialized with IndividualOutcomeEstimator, got ({}) instead."
                            .format(type(estimator)))
        super(OutcomeEvaluator, self).__init__(estimator)
        self._plot_functions.update({"continuous_accuracy": plot_continuous_prediction_accuracy_folds,
                                     "residuals": plot_residual_folds,
                                     "common_support": plot_counterfactual_common_support_folds})

    def _estimator_fit(self, X, a, y):
        """Fit estimator."""
        self.estimator.fit(X=X, a=a, y=y)

    def _estimator_predict(self, X, a):
        """Predict on data."""
        prediction = self.estimator.estimate_individual_outcome(X, a, predict_proba=False)
        # Use predict_probability if possible since it is needed for most evaluations:
        prediction_event_prob = self.estimator.estimate_individual_outcome(X, a, predict_proba=True)

        if prediction_event_prob.columns.tolist() == prediction.columns.tolist():
            # Estimation output for predict_proba=True has same columns as for predict_proba=False.
            # This means either base-learner has no predict_proba/decision_function or problem is not classification.
            # Either way, it means there are no prediction probabilities
            prediction_event_prob = None
        else:
            # predict_proba=True was able to predict probabilities. However,
            # Prediction probability evaluation is only applicable for binary outcome:
            y_values = prediction_event_prob.columns.get_level_values("y").unique()
            #   # Note: on pandas 23.0.0 you could do prediction_event_prob.columns.unique(level='y')
            if y_values.size == 2:
                event_value = y_values.max()  # get the maximal value, assumes binary 0-1 (1: event, 0: non-event)
                # Extract the probability for event:
                prediction_event_prob = prediction_event_prob.xs(key=event_value, axis="columns", level="y")
            else:
                warnings.warn("Multiclass probabilities are not well defined  and supported for evaluation.\n"
                              "Falling back to class predictions.\n"
                              "Plots might be uninformative due to input being classes and not probabilities.")
                prediction_event_prob = None

        fold_prediction = OutcomeEvaluatorPredictions(prediction, prediction_event_prob)
        return fold_prediction

    def score_estimation(self, prediction, X, a_true, y_true, metrics_to_evaluate=None):
        """Scores a prediction against true labels.

        Args:
            prediction (OutcomeEvaluatorPredictions): Prediction on the data.
            X (pd.DataFrame): Covariates.
            a_true (pd.Series): Stratify by - treatment assignment
            y_true: Target variable - outcome.
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives true labels, prediction
                                               and sample_weights (the latter is allowed to be ignored).
                                               If not provided, default are used.

        Returns:
            pd.DataFrame: metric names (keys of `metrics_to_evaluate`) as columns and rows are stratification based
                          on `a_true`, plus unstratified results.
                          For example, binary treatment assignment weill results in three rows: scoring y^0 prediction
                          on the set {y_i | a_i = 0}, y^1 on the set {y_i | a_i = 1} and non-stratified using the
                          factual treatment assignment: y^(a_i) on `y_true`.
        """
        y_is_binary = y_true.nunique() == 2
        treatment_values = np.sort(np.unique(a_true))
        scores = {}
        for treatment_value in treatment_values:
            # Stratify based on treatment assignment:
            current_strata_mask = a_true == treatment_value
            y_true_strata = y_true.loc[current_strata_mask]
            prediction_strata = prediction.prediction.loc[current_strata_mask, treatment_value]
            if y_is_binary:
                prediction_prob_strata = prediction.prediction_event_prob.loc[current_strata_mask, treatment_value]
            else:
                prediction_prob_strata = None

            score = self._score_single(y_true_strata, prediction_strata, prediction_prob_strata,
                                       y_is_binary, metrics_to_evaluate)

            scores[str(treatment_value)] = score

        # Score overall:
        # # Extract prediction on actual treatment
        prediction_strata = robust_lookup(prediction.prediction, a_true)
        if y_is_binary:
            prediction_prob_strata = robust_lookup(prediction.prediction_event_prob, a_true)
        else:
            prediction_prob_strata = None
        score = self._score_single(y_true, prediction_strata, prediction_prob_strata,
                                   y_is_binary, metrics_to_evaluate)
        scores["actual"] = score

        scores = pd.concat(scores, names=["model_strata"], axis="columns").T
        scores = scores.apply(pd.to_numeric, errors="ignore")  # change dtype of each column to numerical if possible.
        return scores

    def _score_single(self, y_true, prediction, prediction_prob, outcome_is_binary, metrics_to_evaluate):
        """Score a single prediction based on whether `y_true` is classification or regression"""
        if outcome_is_binary:
            score = self.score_binary_prediction(y_true, prediction, prediction_prob, metrics_to_evaluate)
        else:
            score = self.score_regression_prediction(y_true, prediction, metrics_to_evaluate)
        # score = pd.DataFrame(score).T
        # score = score.apply(pd.to_numeric, errors="ignore")  # change dtype of each column to numerical if possible.
        return score

    def _get_data_for_plot(self, plot_name, folds_predictions, X, a, y, cv):
        """Retrieve the data needed for each provided plot.
        Plot interfaces are at the plots.py module.

        Args:
            plot_name (str): Plot name.
            folds_predictions (list[OutcomeEvaluatorPredictions]): Predictions for each fold.
            X (pd.DataFrame): Covariates.
            a (pd.Series): Target variable - treatment assignment
            y: *IGNORED*
            cv list[np.ndarray]: Indices (in iloc positions) of each fold.

        Returns:
            tuple: Plot data
        """
        if plot_name in {"continuous_accuracy", "residuals"}:
            folds_predictions_by_actual_treatment = []
            for fold_prediction in folds_predictions:
                if fold_prediction.prediction_event_prob is not None:
                    fold_prediction = fold_prediction.prediction_event_prob
                else:
                    fold_prediction = fold_prediction.prediction
                fold_prediction_by_actual_treatment = robust_lookup(fold_prediction, a[fold_prediction.index])
                folds_predictions_by_actual_treatment.append(fold_prediction_by_actual_treatment)

            return folds_predictions_by_actual_treatment, y, a

        elif plot_name in {'calibration'}:
            folds_predictions = [robust_lookup(prediction.prediction_event_prob, a[prediction.prediction.index])
                                 for prediction in folds_predictions]
            return folds_predictions, y

        elif plot_name in {'roc_curve'}:
            folds_predictions = [robust_lookup(prediction.prediction_event_prob, a[prediction.prediction.index])
                                 for prediction in folds_predictions]
            curve_data = self._calculate_roc_curve_data(folds_predictions, y, a)
            return (curve_data,)

        elif plot_name in {'pr_curve'}:
            folds_predictions = [robust_lookup(prediction.prediction_event_prob, a[prediction.prediction.index])
                                 for prediction in folds_predictions]
            curve_data = self._calculate_pr_curve_data(folds_predictions, y, a)
            return (curve_data,)

        elif plot_name in {'common_support'}:
            if is_vector_binary(y):
                folds_predictions = [prediction.prediction_event_prob for prediction in folds_predictions]
            else:
                folds_predictions = [prediction.prediction for prediction in folds_predictions]
            return folds_predictions, a

        else:
            return None

    def _calculate_curve_data(self, folds_predictions, targets, curve_metric, area_metric, stratify_by=None):
        """Calculate different performance (ROC or PR) curves

        Args:
            folds_predictions (list[pd.Series]): Predictions for each fold.
            targets (pd.Series): True labels
            curve_metric (callable): Performance metric returning 3 output vectors - metric1, metric2 and thresholds.
                                    Where metric1 and metric2 depict the curve when plotted on x-axis and y-axis.
            area_metric (callable): Performance metric of the area under the curve.
            stratify_by (pd.Series): Group assignment to stratify by.

        Returns:
            dict[str, dict[str, list[np.ndarray]]]: Evaluation of the metric for each fold and for each curve.
                One curve for each group level in `stratify_by`.
                On general: {curve_name: {metric1: [evaluation_fold_1, ...]}}.
                For example: {"Treatment=1": {"FPR": [FPR_fold_1, FPR_fold_2, FPR_fold_3]}}
        """
        # folds_targets = [targets.loc[fold_predictions.index] for fold_predictions in folds_predictions]
        # folds_stratify_by = [stratify_by.loc[fold_predictions.index] for fold_predictions in folds_predictions]

        stratify_values = np.sort(np.unique(stratify_by))
        curve_data = {}
        for stratum_level in stratify_values:
            # Slice data for that stratum level across the folds:
            folds_stratum_predictions, folds_stratum_targets = [], []
            for fold_predictions in folds_predictions:
                # Extract fold:
                fold_targets = targets.loc[fold_predictions.index]
                fold_stratify_by = stratify_by.loc[fold_predictions.index]
                # Extract stratum:
                mask = fold_stratify_by == stratum_level
                fold_predictions = fold_predictions.loc[mask]
                fold_targets = fold_targets.loc[mask]
                # Save:
                folds_stratum_predictions.append(fold_predictions)
                folds_stratum_targets.append(fold_targets)

            area_folds, first_ret_folds, second_ret_folds, threshold_folds = \
                self._calculate_performance_curve_data_on_folds(folds_stratum_predictions, folds_stratum_targets, None,
                                                                area_metric, curve_metric)

            curve_data["Treatment={}".format(stratum_level)] = {"first_ret_value": first_ret_folds,
                                                                "second_ret_value": second_ret_folds,
                                                                "Thresholds": threshold_folds, "area": area_folds}
        return curve_data
